// QueueInputStream acts as uni-directional communications pipe between two
// threads, providing an InputStream interface on the reader side. It is
// designed to ensure that writes never block, by queuing all data and
// optionally throwing an exception if the queue reaches a specified size.

package com.tightvnc.decoder.audio;

import java.io.ByteArrayInputStream;
import java.io.InputStream;
import java.io.IOException;
import java.lang.NullPointerException;
import java.util.LinkedList;
import java.util.NoSuchElementException;

public class QueueInputStream extends InputStream {

  protected LinkedList queue;
  protected long maxSize = 0;
  protected long size = 0;
  protected boolean closed = false;

  // No queue size limit
  public QueueInputStream() {
    queue = new LinkedList();
  }

  public QueueInputStream(long maxQueuedBytes) {
    queue = new LinkedList();
    maxSize = maxQueuedBytes;
  }

  public void write(byte[] data) throws IOException {
    if (closed) { throw new IOException("Stream has been closed"); }

    synchronized (queue) {
      if (maxSize != 0 && size + data.length > maxSize) {
        throw new IOException("Queue is full");
      }
      size += data.length;
      queue.addLast(data);
      queue.notifyAll();
    }
  }

  public long getMaxSize() {
    return maxSize;
  }

  public long getQueuedSize() {
    return size;
  }

  protected InputStream current = null;

  protected InputStream getCurrentStream() throws IOException {
    if (current == null) {
      return getNextStream();
    }
    return current;
  }

  protected InputStream getNextStream() throws IOException {
    synchronized (queue) {
      while(true) {
        try {
          byte[] data = (byte[])(queue.removeFirst());
          size -= data.length;
          current = new ByteArrayInputStream(data);
          break;
        } catch (NoSuchElementException nsee) {
          if (closed) {
            queue = null;
            current = null;
            throw new IOException("Stream has been closed");
          }
          try { queue.wait(); } catch (InterruptedException ie) {}
        }
      }
    }
    return current;
  }

  public int read() throws IOException {
    int data;
    try {
      data = getCurrentStream().read();
      if (data == -1) { data = getNextStream().read(); }
    } catch (IOException e) { data = -1; }
    return data;
  }

  public int read(byte[] data) throws IOException {
    if (data == null) { throw new NullPointerException(); }

    int bytes;
    try {
      bytes = getCurrentStream().read(data);
      if (bytes == -1) { bytes = getNextStream().read(data); }
    } catch (IOException e) { bytes = -1; }
    return bytes;
  }

  public int read(byte[] data, int offset, int length) throws IOException {
    if (data == null) { throw new NullPointerException(); }
    else if (offset < 0 || length < 0 || length > data.length - offset) {
      throw new IndexOutOfBoundsException();
    }

    int bytes;
    try {
      bytes = getCurrentStream().read(data, offset, length);
      if (bytes == -1) { bytes = getNextStream().read(data, offset, length); }
    } catch (IOException e) { bytes = -1; }
    return bytes;
  }

  public int available() throws IOException {
    if (closed || (current == null & queue.isEmpty())) {
      return 0;
    }
    int bytes = 0;
    if (current != null) {
      bytes = getCurrentStream().available();
    }
    if (bytes == 0 && !queue.isEmpty()) {
      bytes = getNextStream().available();
    }
    return bytes;
  }

  public void close() throws IOException {
    closed = true;
    synchronized (queue) {
      queue.notifyAll();
    }
  }

}
